package net.knarcraft.placeholdersigns.listener;

import net.knarcraft.knarlib.formatting.StringFormatter;
import net.knarcraft.knarlib.formatting.StringReplacer;
import net.knarcraft.knarlib.property.ColorConversion;
import net.knarcraft.knarlib.util.ColorHelper;
import net.knarcraft.placeholdersigns.PlaceholderSigns;
import net.knarcraft.placeholdersigns.config.PlaceholderSignMessage;
import net.knarcraft.placeholdersigns.container.SignLineChangeRequest;
import net.knarcraft.placeholdersigns.handler.PlaceholderSignRequestHandler;
import net.md_5.bungee.api.ChatColor;
import org.bukkit.Bukkit;
import org.bukkit.DyeColor;
import org.bukkit.block.Sign;
import org.bukkit.block.sign.Side;
import org.bukkit.block.sign.SignSide;
import org.bukkit.entity.Player;
import org.bukkit.event.EventHandler;
import org.bukkit.event.EventPriority;
import org.bukkit.event.Listener;
import org.bukkit.event.block.SignChangeEvent;
import org.bukkit.event.player.PlayerInteractEvent;
import org.jetbrains.annotations.NotNull;

import java.util.Objects;

/**
 * A listener for placeholder signs being clicked
 */
public class SignClickListener implements Listener {

    @EventHandler(priority = EventPriority.MONITOR)
    public void onSignClick(@NotNull PlayerInteractEvent event) {
        // Ignore if not a clicked sign
        if (!event.hasBlock() || event.getClickedBlock() == null ||
                !(event.getClickedBlock().getState() instanceof Sign sign)) {
            return;
        }

        Player player = event.getPlayer();
        PlaceholderSignRequestHandler requestHandler = PlaceholderSigns.getInstance().getRequestHandler();
        Boolean hasSignViewRequest = requestHandler.getSignViewRequest(player);

        if (hasSignViewRequest != null) {
            printSign(sign, player, hasSignViewRequest);
            // Cancel the event to prevent vanilla behavior
            event.setCancelled(true);
            return;
        }

        if (sign.isWaxed() && !player.hasPermission("placeholdersigns.edit.bypass-waxed")) {
            PlaceholderSigns.getInstance().getStringFormatter().displayErrorMessage(player,
                    PlaceholderSignMessage.ERROR_WAXED_NO_PERMISSION);
            return;
        }

        // Check if the player has run the /editSign command
        SignLineChangeRequest request = requestHandler.getSignChangeRequest(player);
        if (request != null) {
            SignSide standingOn = sign.getTargetSide(player);
            Side side = sign.getSide(Side.FRONT).equals(standingOn) ? Side.FRONT : Side.BACK;
            doSignChange(sign, request, event, player, side);
        }
    }

    /**
     * Prints the current contents of a sign to a player
     *
     * @param sign   <p>The sign to print</p>
     * @param player <p>The player to display the contents to</p>
     * @param raw    <p>Whether to get the raw text with used formatting codes</p>
     */
    private void printSign(@NotNull Sign sign, @NotNull Player player, boolean raw) {
        SignSide front = sign.getSide(Side.FRONT);
        SignSide back = sign.getSide(Side.BACK);
        String frontLines = getSignText(front, raw);
        String backLines = getSignText(back, raw);
        StringFormatter stringFormatter = PlaceholderSigns.getInstance().getStringFormatter();
        StringReplacer replacer = new StringReplacer(stringFormatter.getUnformattedColoredMessage(
                PlaceholderSignMessage.SUCCESS_SIGN_CONTENTS));
        replacer.add("{frontLines}", frontLines);
        replacer.add("{backLines}", backLines);
        replacer.add("{frontDye}", getDye(front));
        replacer.add("{frontGlow}", getGlow(front));
        replacer.add("{backDye}", getDye(back));
        replacer.add("{backGlow}", getGlow(back));
        player.sendMessage(replacer.replace());
    }

    /**
     * Gets a description of the glow of a sign
     *
     * @param signSide <p>The sign side to get the dye of</p>
     * @return <p>The description of the applied dye</p>
     */
    @NotNull
    private String getGlow(@NotNull SignSide signSide) {
        boolean glowing = signSide.isGlowingText();
        StringFormatter stringFormatter = PlaceholderSigns.getInstance().getStringFormatter();
        if (glowing) {
            return stringFormatter.getUnformattedColoredMessage(PlaceholderSignMessage.GLOWING_CONFIRM);
        } else {
            return stringFormatter.getUnformattedColoredMessage(PlaceholderSignMessage.GLOWING_DENY);
        }
    }

    /**
     * Gets a description of a dye applied to a sign
     *
     * @param signSide <p>The sign side to get the dye of</p>
     * @return <p>The description of the applied dye</p>
     */
    @NotNull
    private String getDye(@NotNull SignSide signSide) {
        DyeColor dyeColor = signSide.getColor();
        if (dyeColor == null) {
            return "None";
        }

        ChatColor color = ColorHelper.fromColor(dyeColor.getColor());
        return color + signSide.getColor().name();
    }

    /**
     * Gets text from a sign side, and appends it to the given string builder
     *
     * @param signSide <p>The sign side to get text from</p>
     * @param raw      <p>Whether to get the raw text with used formatting codes</p>
     */
    @NotNull
    private String getSignText(@NotNull SignSide signSide, boolean raw) {
        StringBuilder output = new StringBuilder();
        for (int i = 0; i < 4; i++) {
            output.append(i + 1).append(". ");

            String line = signSide.getLine(i);
            if (raw) {
                output.append(line.replace(ChatColor.COLOR_CHAR, '&'));
            } else {
                output.append(line);
            }
            output.append("\n");
        }
        return output.toString();
    }

    /**
     * Perform the changing of a sign according to the given sign change request
     *
     * @param sign    <p>The sign to be changed</p>
     * @param request <p>The sign line change request to perform</p>
     * @param event   <p>The interaction event that triggered this</p>
     * @param player  <p>The player triggering the sign change</p>
     */
    private void doSignChange(@NotNull Sign sign, @NotNull SignLineChangeRequest request,
                              @NotNull PlayerInteractEvent event, @NotNull Player player, @NotNull Side side) {
        // Cancel the event to prevent vanilla behavior
        event.setCancelled(true);

        SignSide signSide = sign.getSide(side);
        String[] lines = signSide.getLines();

        lines[request.line()] = request.text();

        // Run the sign change event to allow protection plugins to cancel, and allow the sign text listener to trigger
        SignChangeEvent changeEvent = new SignChangeEvent(Objects.requireNonNull(event.getClickedBlock()),
                player, lines, side);
        Bukkit.getPluginManager().callEvent(changeEvent);
        if (changeEvent.isCancelled()) {
            return;
        }

        // Update the sign with the new text
        String[] finalLines = changeEvent.getLines();
        for (int i = 0; i < finalLines.length; i++) {
            signSide.setLine(i, ColorHelper.translateColorCodes(finalLines[i], ColorConversion.RGB));
        }
        sign.update();

        PlaceholderSigns.getInstance().getStringFormatter().displaySuccessMessage(player,
                PlaceholderSignMessage.SUCCESS_SIGN_CHANGED);
    }

}
